# Section 18.2 Question 2

def gcd(a, b):
    """Return the greatest common divisor of a and b."""
    d = min(a, b)
    while d > 1:
        if a % d == 0 and b % d == 0:
            return d
        d -= 1
    return 1

# Section 18.2 Question 3

def gcd(a, b):
    """Return the greatest common divisor of a and b."""
    if a > b:
        return gcd(a - b, b)
    elif a < b:
        return gcd(a, b - a)
    else:
        return a     # or return b

# Section 18.2 Question 4

def gcd(a, b):
    while a != b:
        if a > b:
            a -= b
        else:
            b -= a
    return a           # or return b

if __name__ == '__main__':
    # print(gcd(18289894500228625200, 14814814692))
    # Runs "forever"
    pass

def gcd(a, b):
    """Return the greatest common divisor of a and b."""
    while a > 0:
        b, a = a, b % a
    return b

if __name__ == '__main__':  # Run only as a stand-alone program,
                            # not when imported into other programs
    print(gcd(18289894500228625200, 14814814692))

# Section 18.2 Question 7

def solve_diophantine(a, b):
    """Return x and y such that ax + by = 1."""
    y = 0
    while y < a and b*y % a != 1:
        y += 1
    if b*y % a == 1:
        x = (1 - b*y)//a
        return (x, y)
    else:
        return None

if __name__ == '__main__':
    print(solve_diophantine(7, -5))

# Section 18.2 Question 8

def solve_diophantine(a, b, from_x, to_x):
    """Return x and y such that ax + by = 1 and from_x <= x <= to_x."""
    x = from_x
    while x <= to_x and a*x % abs(b) != 1:
        x += 1
    if a*x % abs(b) == 1:
        y = (1 - a*x)//b
        return (x, y)
    else:
        return None

if __name__ == '__main__':
    print(solve_diophantine(7, -5, 20, 30))

# Section 18.2 Question 9

def solve_diophantine(a, b, c):
    """Return x, y such that ax + by = c."""
    if abs(a) > abs(b):
        t = solve_diophantine(b, a, c)
        if t is None:
            return None
        return t[1], t[0]

    if a == 0:
        if c % b == 0:
            return 0, c // b
        else:
            return None
    
    elif b % a == 0:
        if c % a == 0:
            return c//a, 0
        else:
            return None

    k, d = b//a, b % a
    t = solve_diophantine(d, a, c)
    if t is None:
        return None
    x1, y1 = t
    return y1 - k*x1, x1

if __name__ == '__main__':
    print(solve_diophantine(28, -20, 8))

# Section 18.3 Question 3

def print_factorization(n):
    print(n, end='')
    separator = ' ='
    d = 2
    while n > 1:
        if n % d == 0:   # the smallest divisor > 1 is always a prime
            print(separator, d, end='')
            separator = ' *'
            n //= d
        else:
            d += 1
    print()

if __name__ == '__main__':
    n = int(input('Enter a positive integer: '))
    print_factorization(n)

# Section 7.4 Example 5

from math import sqrt

def is_prime(n):
    """Return True if n is a prime number; otherwise return False."""
    if n < 2:
        return False
    d = 2
    while d <= round(sqrt(n)):  # if n is divisible by k, then n is
                                #   also divisible by n/k and either
                                #   k or n/k must be below sqrt(n)
        if n % d == 0:
            return False
        d += 1
    return True

# Section 18.3 Question 4

def find_first_composite():
    """Returns the first composite = p_1*p_2*...*p_n + 1."""
    n = 1
    p = 2
    while p <= 1000:
        if is_prime(p):  # is_prime is described in Section 7.4
            n *= p
            if not is_prime(n+1):
                return n+1
        p += 1
    if p > 1000:
        return None

if __name__ == '__main__':
    n = find_first_composite()
    print_factorization(n-1)
    print_factorization(n)

# Section 18.3 Question 5

if __name__ == '__main__':
    n = 1
    while is_prime(n*n - n + 41):
        n += 1
    print('n =', n)

# Section 18.3 Question 8

if __name__ == '__main__':
    goldbach = True

    for n in range(6, 101, 2): # obviously true for n = 4
        found2primes = False
        for p in range(3, n//2+1, 2):
            if is_prime(p) and is_prime(n - p):
                found2primes = True
                break
        if not found2primes:
            print(n,  'is not a sum of two primes!')
            goldbach = False

    if goldbach:
        print("Goldbach's conjecture is true for n = 4, ..., 100")

# Section 18.4 Question 15

def pow_mod(a, n, p):
    """Return a^n % p."""
    a %= p
    if n == 0:
        return 1
    elif n == 1:
        return a
    else:
        n, r = divmod(n, 2)
        y = pow_mod(a, n, p)
        y = y*y % p
        if r == 1:
            y = y*a % p
        return y

from random import randint

def is_likely_prime(p):
    """Return True if p is a likely prime as indicated by
       several trials with Fermat's Little Theorem; otherwise returns False.
    """
    if p < 10:
        return p == 2 or p == 3 or p == 5 or p == 7
    low = 3
    hi = p//2
    num_tries = min(hi - low, 100)
    count = 0
    while count < num_tries:
        a = randint(low, hi)
        if pow_mod(a, p-1, p) != 1:
            return False
        count += 1
    return True

if __name__ == '__main__':
    p = 1000001
    while not is_likely_prime(p):
        p += 2
    print(p)

# Section 18.5 Question 4

if __name__ == '__main__':
    p = 170141183460469231731687303715884105727
    a = 618970019642690137449562111
    r = 5
    print(pow_mod(r, a, p))
